package index.MyMtree;

import java.util.Arrays;

import index.structures.IndexMyMtree;

/**
 * M-tree with SS-Tree split algorithm
 * @author amartin
 *
 */
public final class Sphere {
	int[] center;
	int radius;
	public Sphere[] children = null;
	int size;
	Sphere parent;
	final int TID;
	
	private static final SSTreeSplitHandler SPLIT_HANDLER = new SSTreeSplitHandler();
	public static final int INNER_NODE = -1;
	
	/**
	 * Point Constructor
	 */
	public Sphere(final int[] _point, final int _TID, final Sphere _parent){
		int[] clone = new int[_point.length];
		System.arraycopy(_point, 0, clone, 0, _point.length);
		this.center = clone;
		this.radius = 0;
		this.size	= 1;
		this.TID	= _TID;
		this.parent = _parent;
	}
	
	/**
	 * Inner node (empty) constructor, for split-handling
	 */
	public Sphere(final int SPLIT_SIZE, final Sphere _parent){
		this.center = new int[IndexMyMtree.tree.DIM];
		this.radius = 0;
		this.size 	= 0;
		this.TID	= INNER_NODE;
		this.children = new Sphere[SPLIT_SIZE];
		this.parent	= _parent;
	}
	
	/**
	 * Root Constructor
	 */
	private Sphere(){
		this.radius = 0;
		this.size	= 0;
		this.TID	= INNER_NODE;
		this.parent = null;
		this.children = new Sphere[IndexMyMtree.tree.SPLIT_SIZE];
		this.center = new int[IndexMyMtree.tree.DIM];
	}
	
	public static Sphere getRoot(){
		return new Sphere();
	}
	
	public void add(final int[] point, final int TID, final int SPLIT_SIZE){
		//(1) expand borders if necessary, recall (0,..,0) is in the UPPER_LEFT
		Sphere toAdd = new Sphere(point, TID, this);
		this.children[size()] = toAdd;
		this.size++;
			
		if(size()==SPLIT_SIZE){
			SPLIT_HANDLER.split(this);
		}else{
			// Check whether we need to change the radius
			this.checkForAdjust();
		}
	}
	
	int innerDistManhattan(final int[] point){
		int dist = 0;
		for(int dim=0;dim<point.length;dim++){
			dist += (point[dim]>center[dim]) ? point[dim]-center[dim] : center[dim]-point[dim];//abs()
		}
		return dist;
	}
	int innerDistEuclid(final int[] point){
		
		int dist = 0;
		for(int dim=0;dim<point.length;dim++){
			dist += (point[dim]-center[dim]) * (point[dim]-center[dim]);
		}
		return (int) Math.sqrt(dist);
	}
	int innerDist(final int[] point){
		return (int) IndexMyMtree.tree.getMetric().calculateDistance(center, point);
	}
	
	public Sphere getLeaf(final int[] _dataset){
		if(isLeaf())
			return this;
		else{
			//which leaf to return?
			final InsertPunishement[] punishements = new InsertPunishement[this.size()];
			for(int leaf=0;leaf<size();leaf++){
				punishements[leaf] = punishement(this, leaf, _dataset);
			}
			//sort so that Sphere with least overlap is first, resolve ties by min deltaVolume
			Arrays.sort(punishements);
			
			if(children[0].isLeaf())//if next level contains leaves terminate
				return punishements[0].sphere;
			else
				return punishements[0].sphere.getLeaf(_dataset);
		}
	}

	private InsertPunishement punishement(final Sphere parent, final int indexChild, final int[] _dataset) {
		
		final Sphere toTest = parent.children[indexChild];
		final int[] coordinates = new int[toTest.center.length];
		System.arraycopy(toTest.center, 0, coordinates, 0, toTest.center.length);
		final int SIZE = toTest.size();
	
		///////////////////////////BEGIN Radius///////////////////////
		// Compute potential new center
		int delta;
		for(int dim=0;dim<toTest.center.length;dim++){
			delta = _dataset[dim]-coordinates[dim];//(1) Order is important!
			delta /= (SIZE+1);
			coordinates[dim] += delta;//(1) Therefore.
		}
		
		// Check whether we need to change the radius
		int dist2center = -1;
		int newRadius = -1;
		for(int child=0;child<SIZE;child++){
			Sphere temp = toTest.children[child];
			dist2center = toTest.innerDist(temp.center);
			if(dist2center>newRadius)
				newRadius = dist2center;
		}
		///////////////////////////END Radius///////////////////////
		
		///////////////////////////Begin Overlap////////////////////
		long overlap = 0;
		
		for(int sphere=0; sphere<parent.size(); sphere++){
			if(sphere==indexChild)
				continue;
			final Sphere sibling = parent.children[sphere];
			int dist2sibling = sibling.innerDist(coordinates);
			dist2sibling -= newRadius;
			dist2sibling -= sibling.radius;
			if(dist2sibling>0)
				overlap+=dist2sibling;
		}
		
		return new InsertPunishement(overlap, newRadius, toTest);
	}

	public int size(){
		return this.size;
	}
	
	public final boolean isPoint()	{	return children == null;		}
	public final boolean isLeaf()	{	
		if (isPoint())
			return false;
		if (isRoot()){
			if(children[0]==null)
				return true;
			return children[0].children == null;
		}
		return children[0].children == null;		
	}
	public final boolean isRoot()	{	return parent == null;	}//has no parent
	
	/**
	 * Simple class helping to find the sphere where to insert a new point.
	 * @author amartin
	 *
	 */
	private final class InsertPunishement  implements Comparable<InsertPunishement>{
		final long overlap;
		final int newRadius;
		final Sphere sphere;

		public InsertPunishement(final long overlap, final  int newRadius, final Sphere sphere) {
			this.overlap 		= overlap;
			this.newRadius  	= newRadius;
			this.sphere 		= sphere;
		}
		
		public final int compareTo(InsertPunishement compare){
			return (int)( (this.overlap!=compare.overlap) ? this.overlap-compare.overlap : this.newRadius-compare.newRadius);
		}
	}

	public void computeCenterAndRadius() {
		//compute center dimension by dimension
		final int SIZE = size();
		int tempCoordinate;
		
		for(int dim=0;dim<center.length;dim++){
			tempCoordinate = 0;
			for(int child=0;child<SIZE;child++){
				tempCoordinate += children[child].center[dim]; 
			}
			tempCoordinate /= SIZE;
			this.center[dim] = tempCoordinate;
		}
		
		//compute required radius
		int dist2center = 0;
		int radius = 0;
		
		for(int child=0;child<SIZE;child++){
			dist2center = innerDist(children[child].center);
			int totalDist = dist2center +children[child].radius;
			if(radius<totalDist){
				radius = totalDist;
			}
		}
		this.radius = radius;
	}
	
	/**
	 * It is possible that, due to splitting, we need to adjust parents, here we check and adjust if necessary
	 * @param parent
	 */
	public void checkForAdjust() {
		boolean requireAdjust = false;
		
		for(int child=0;child<size();child++){
			if(!this.contains(this.children[child])){
				requireAdjust = true;
				break;
			}
		}
		
		
		if(requireAdjust){
			computeCenterAndRadius();
			if(parent!=null)
				parent.checkForAdjust();
		}
	}

	/////////////////////////////Query stuff/////////////////////////
	
	public int[] getExact(final int[] _datasetToSearch) {
		if(isLeaf()){
			//check each point
			for(int i=0;i<size();i++){
				Sphere point = this.children[i];
				boolean identic = true;
				//HDD access or Access penalty
				//final int[] key = index.structures.IndexMySSTree.tree.getDataset(point.TID);
				final int[] key = point.center;
				
				for(int dim=0;dim<point.center.length;dim++){
					if(key[dim]!=_datasetToSearch[dim]){
						identic = false;
						break;
					}
				}
				if(identic){
					return point.center;
				}
			}
		}else{
			for(int i=0;i<size();i++){
				Sphere sphere = this.children[i];
				if(sphere.covers(_datasetToSearch)){
					int[] result = sphere.getExact(_datasetToSearch);
					if(result != null)	
						return result; 
				}
			}
		}
		return null;//not found in this branch
	}
	
	public void getNN(Neighbors n){
		if(isLeaf()){
			//try update of nearest neighbor list
			Sphere point;
			for(int i=0;i<size();i++){
				point = this.children[i];
				n.update(point);
			}
		}else{
			//consider every sub tree that may contain points with smeller dist than current may dist
			SphereDists[] subTrees = new SphereDists[size()];	
			Sphere subTree;
			int distEstimation;
			
			for(int i=0;i<size();i++){
				subTree = this.children[i];
				// dist of both centers
				distEstimation = subTree.innerDist(n.query);
				//so, in the worst case there is a point on the straight 
				//line between both centers with dist radius from sub tree center
				//XXX Note, this works only iff triangle inequality for applied metric (and space) hold 
				distEstimation -= subTree.radius;
				if(distEstimation<0)
					distEstimation = 0;
				subTrees[i] = new SphereDists(subTree, distEstimation);
			}
			Arrays.sort(subTrees);
			
			for(int i=0; i<subTrees.length;i++){
				SphereDists sd = subTrees[i];
				if(sd.distEstimation<n.maxDist){//XXX long vs. double
					sd.sphere.getNN(n);
				}
			}
		}
	}

	private boolean covers(int[] _datasetToSearch) {
		int dist2center = innerDist(_datasetToSearch);
		return (dist2center-this.radius)<=0;
	}
	
	///////////////////////////For debug/////////////////////////
	public String toString(){
		String ret = "";
		if(isRoot()){
			ret+="Root ";
		}
		if(isLeaf()){
			ret+="Leaf ";
		}
		if(isPoint()){
			ret+="Point "+TID+" ";
		}
		ret+="c={";
		for(int dim=0;dim<center.length;dim++){
			ret+=center[dim]+" ";
		}
		ret+="} r="+radius;
		return ret;
	}

	public boolean contains(Sphere child) {
		return contains(child.center, child.radius);
	}
	
	public boolean contains(final int[] _point, final int _radius) {
		int dist2center = this.innerDist(_point);
		dist2center += _radius;
		return (dist2center<=this.radius);
	}
	
	public final class SphereDists implements Comparable<SphereDists>{
		final long distEstimation;
		final Sphere sphere;
		
		SphereDists(final Sphere sphere, final int distEstimation){
			this.sphere = sphere;
			this.distEstimation =  distEstimation;
		}

		@Override
		public int compareTo(SphereDists o) {
			return (distEstimation<o.distEstimation) ? 1 : -1;
		}
		
		public String toString(){
			return "" + distEstimation + sphere;
		}
	}
}
